package customnode;

import javax.vecmath.Vector3f;
import javax.vecmath.Point3f;

import javax.media.j3d.GeometryArray;
import com.sun.j3d.utils.geometry.GeometryInfo;

import java.util.Collections;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.List;
import java.util.Comparator;
import java.util.PriorityQueue;

public class DecimateMesh {

	private List<Point3f> mesh;

	private List<Triangle> faces;
	private PriorityQueue<Vertex> queue;
	private Map<Point3f, Vertex> vertices;

	public DecimateMesh(List<Point3f> mesh) {
		this.mesh = mesh;
	}

	private int iter = 0;

	public void decimate() {
		init();
		queue = new PriorityQueue<Vertex> (
				vertices.size(), new VertexComparator());
		for(Point3f p : vertices.keySet()) {
			Vertex v = vertices.get(p);
			v.calcSumOfNormalDeviation();
			v.surroundingPolygon();
			queue.offer(v);
		}
		while(!queue.isEmpty()) {
			int dbg = 1;
			iter++;
			Vertex v = queue.poll();
			System.out.println(v + " = " + v.sumOfNormalDeviation);
			if(v.sumOfNormalDeviation > 0.2)
				break;
			if(iter == 1)
				showVertex(v, "before " + iter);
			v.retriangulate();
			if(iter == 1)
				showVertex(v, "after " + iter);
			vertices.remove(v.point);
			if(iter == dbg)
				break;
		}
		mesh.clear();
		for(Triangle t : faces) {
			mesh.add(t.v1.point);
			mesh.add(t.v2.point);
			mesh.add(t.v3.point);
		}
	}

	private void init() {
		faces = new ArrayList<Triangle>();
		vertices = new HashMap<Point3f, Vertex>();
		for(int i = 0; i < mesh.size(); i += 3) {
			Vertex v1 = newVertex(mesh.get(i));
			Vertex v2 = newVertex(mesh.get(i + 1));
			Vertex v3 = newVertex(mesh.get(i + 2));
			Triangle t = new Triangle(v1, v2, v3);
			v1.addTriangle(t);
			v2.addTriangle(t);
			v3.addTriangle(t);
			faces.add(t);
		}
	}

	private Vertex newVertex(Point3f p) {
		if(vertices.containsKey(p))
			return vertices.get(p);
		Vertex v = new Vertex(p);
		vertices.put(p, v);
		return v;
	}

	private static class VertexComparator implements Comparator<Vertex> {
		public int compare(Vertex v1, Vertex v2) {
			if(v1.sumOfNormalDeviation < v2.sumOfNormalDeviation)
				return -1;
			if(v1.sumOfNormalDeviation > v2.sumOfNormalDeviation)
				return +1;
			return 0;
		}
	}

	private static class Triangle {
		private Vertex v1, v2, v3;

		public Triangle(Vertex v1, Vertex v2, Vertex v3) {
			this.v1 = v1;
			this.v2 = v2;
			this.v3 = v3;
		}

		public float area() {
			Vector3f vc1 = new Vector3f();
			Vector3f vc2 = new Vector3f();
			vc1.sub(v2.point, v1.point);
			vc2.sub(v3.point, v1.point);
			vc1.cross(vc1, vc2);
			return vc1.length();
		}

// 		public boolean equals(Object o) {
// 			Triangle t2 = (Triangle)o;
// 			Set<Point3f> p = new HashSet<Point3f>();
// 			p.add(v1.point);
// 			p.add(v2.point);
// 			p.add(v3.point);
// 			p.remove(t2.v1.point);
// 			p.remove(t2.v2.point);
// 			p.remove(t2.v3.point);
// 			return p.isEmpty();
// 		}

		public String toString() {
			return "Triangle: " + v1.point + v2.point + v3.point;
		}
	}

	private class Vertex implements Comparable<Vertex> {
		private Point3f point;
		private Set<Triangle> triangles;
		private float sumOfNormalDeviation = 0;

		public Vertex(Point3f p) {
			point = p;
			triangles = new HashSet<Triangle>();
		}

		public void addTriangle(Triangle t) {
			triangles.add(t);
		}

		public void delTriangle(Triangle t) {
			triangles.remove(t);
		}

		public Point3f[] surroundingPolygon() {
			if(triangles.size() < 3) {
				System.out.println(triangles);
				throw new RuntimeException("Polygon with 2 vertices");
			}
			Map<Vertex, Vertex> edges = new HashMap<Vertex, Vertex>();
			for(Triangle t : triangles) {
				if(t.v1.point.equals(point))
					edges.put(t.v2, t.v3);
				if(t.v2.point.equals(point))
					edges.put(t.v3, t.v1);
				if(t.v3.point.equals(point))
					edges.put(t.v1, t.v2);
			}
			Point3f[] coords = new Point3f[triangles.size()];
			Vertex next = Collections.min(edges.keySet());
			int i = 0;
			coords[i++] = next.point;
			while(!edges.isEmpty()) {
Vertex tmp = next;
				next = edges.remove(next);
				if(next == null) {
edges.clear();
for(Triangle t : triangles) {
	if(t.v1.point.equals(point))
		edges.put(t.v2, t.v3);
	if(t.v2.point.equals(point))
		edges.put(t.v3, t.v1);
	if(t.v3.point.equals(point))
		edges.put(t.v1, t.v2);
}
for(Vertex key : edges.keySet())
	System.out.println(key + " -> " + edges.get(key));
System.out.println("did not find " + tmp);
showVertex(this);
					throw new RuntimeException("not a simple vertex");
				}
				if(next.point.equals(coords[0]))
					break;
try {
				coords[i++] = next.point;
} catch(RuntimeException e) {
edges.clear();
for(Triangle t : triangles) {
	if(t.v1.point.equals(point))
		edges.put(t.v2, t.v3);
	if(t.v2.point.equals(point))
		edges.put(t.v3, t.v1);
	if(t.v3.point.equals(point))
		edges.put(t.v1, t.v2);
}
for(Vertex key : edges.keySet())
	System.out.println(key + " -> " + edges.get(key));
System.out.println("did not find " + tmp);
showVertex(this);
throw e;
}
			}
			if(!next.point.equals(coords[0]))
				throw new RuntimeException("not a simple vertex");
			if(!edges.isEmpty()) {
edges.clear();
for(Triangle t : triangles) {
	if(t.v1.point.equals(point))
		edges.put(t.v2, t.v3);
	if(t.v2.point.equals(point))
		edges.put(t.v3, t.v1);
	if(t.v3.point.equals(point))
		edges.put(t.v1, t.v2);
}
for(Vertex key : edges.keySet())
	System.out.println(key + " -> " + edges.get(key));
System.out.println("triangles.size = " + triangles.size());
showVertex(this);
				throw new RuntimeException("not a simple vertex");
			}
			return rearrange(coords);
		}

		private Point3f[] rearrange(Point3f[] vertices) {
if(iter == 1) {
	System.out.println("before rearrange");
	for(Point3f p : vertices)
		System.out.println(p);
}
			int i = 0;
			int l = vertices.length;
			boolean[] online = new boolean[vertices.length];
			for(i = 0; i < l; i++) {
				online[i] = pointOnLine(vertices[i],
						vertices[(i - 1 + l) % l],
						vertices[(i + 1) % l]);
			}
			// we are looking for a point whose both neighbors are
			// not online. (It does not matter if the point itself
			// is online)
			boolean success = false;
			for(i = 0; i < l; i++) {
				if(!online[(i - 1 + l) % l] && !online[(i + 1) % l]) {
					success = true;
					break;
				}
			}
			if(!success) {
				System.out.println("Could not re-arrange");
				return null;
			}
			Point3f[] copy = new Point3f[l];
			System.arraycopy(vertices, i, copy, 0, l - i);
			System.arraycopy(vertices, 0, copy, l - i, i);
if(iter == 1) {
	System.out.println("after rearrange");
	for(Point3f p : copy)
		System.out.println(p);
}
			return copy;
		}

		private boolean pointOnLine(Point3f p, Point3f ls, Point3f le) {
			float tx = (p.x - ls.x) / (le.x - ls.x);
			float ty = (p.y - ls.y) / (le.y - ls.y);
			float tz = (p.z - ls.z) / (le.z - ls.z);
			boolean b =  fleq(tx, ty) && fleq(ty, tz) && tx >= 0 && tx <= 1;
			if(b)
				System.out.println("found point on line");
			return b;
		}

		private boolean fleq(float f1, float f2) {
			return Math.abs(f1 - f2) < 0.001f;
		}

		public void retriangulate() {
			Point3f[] coords = surroundingPolygon();

			// retriangulate
			GeometryInfo gi = new GeometryInfo(
							GeometryInfo.POLYGON_ARRAY);
			gi.setCoordinates(coords);
			gi.setStripCounts(new int[] {coords.length});
			gi.convertToIndexedTriangles();
			gi.unindexify();
			GeometryArray ga = gi.getGeometryArray();
			int n = ga.getVertexCount();
			Point3f[] newCoords = new Point3f[n];
			for(int i = 0; i < newCoords.length; i++)
				newCoords[i] = new Point3f();
			ga.getCoordinates(0, newCoords);

			// make new triangles
			Set<Triangle> newTriangles = new HashSet<Triangle>();
			for(int i = 0; i < n; i += 3) {
				Vertex v1 = vertices.get(newCoords[i]);
				Vertex v2 = vertices.get(newCoords[i + 1]);
				Vertex v3 = vertices.get(newCoords[i + 2]);
				if(v1 == null || v2 == null || v3 == null)
					throw new RuntimeException("created new vertex");
				Triangle tri = new Triangle(v1, v2, v3);
				if(tri.area() < 0.001f)
					continue;
				newTriangles.add(tri);
			}

			// collect neighbor vertices
			Set<Vertex> neighbors = new HashSet<Vertex>();
			for(Triangle t : triangles) {
				neighbors.add(t.v1);
				neighbors.add(t.v2);
				neighbors.add(t.v3);
			}
			neighbors.remove(this);
			if(neighbors.size() > triangles.size())
				throw new RuntimeException();

			// remove neighbors from queue
			for(Vertex ne : neighbors)
				queue.remove(ne);

			// remove old triangles
			for(Triangle t : triangles) {
				faces.remove(t);
				for(Vertex ne : neighbors)
					ne.delTriangle(t);
			}
			// add new triangles
			faces.addAll(newTriangles);
			for(Triangle t : newTriangles) {
				t.v1.addTriangle(t);
				t.v2.addTriangle(t);
				t.v3.addTriangle(t);
			}

			// recalculate costs
			for(Vertex ne : neighbors) {
				ne.calcSumOfNormalDeviation();
				try {
					ne.surroundingPolygon();
				} catch(RuntimeException e) {
					System.out.print(e.getMessage());
					System.out.println("old triangles " + triangles);
					System.out.println("new triangles " + newTriangles);
					throw e;
				}
				queue.add(ne);
			}

			// update self
			triangles.clear();
			triangles.addAll(newTriangles);
		}

		public void calcSumOfNormalDeviation() {
			Vector3f n = new Vector3f();
			Vector3f v1 = new Vector3f();
			Vector3f v2 = new Vector3f();
			Vector3f[] normals = new Vector3f[triangles.size()];
			// average normal vector
			int i = 0;
			for(Triangle t : triangles) {
				v1.sub(t.v2.point, t.v1.point);
				v2.sub(t.v3.point, t.v1.point);
				v1.cross(v1, v2);
				n.add(v1);
				v1.normalize();
				normals[i++] = new Vector3f(v1);
			}
			n.normalize();
			// get the maximum angle between normals and avg normal
			this.sumOfNormalDeviation = 0;
			for(Vector3f no : normals) {
				double a = Math.acos(no.dot(n));
				if(a > sumOfNormalDeviation)
					sumOfNormalDeviation = (float)a;
			}
		}

		public boolean equals(Object o) {
			return point.equals(((Vertex)o).point);
		}

		public int hashCode() {
			return point.hashCode();
		}

		public String toString() {
			return point.toString();
		}

		public int compareTo(Vertex o) {
			if(point.z != o.point.z) 
				return (int)Math.signum(point.z - o.point.z);
			if(point.y != o.point.y)
				return (int)Math.signum(point.y - o.point.y);
			if(point.x != o.point.x)
				return (int)Math.signum(point.x - o.point.x);
			return 0;
		}
	}

	public void showVertex(Vertex v) {
		showVertex(v, "bla");
	}

	public void showVertex(Vertex v, String bla) {
		List<Point3f> m = new ArrayList<Point3f>();
		for(Triangle t : v.triangles) {
			m.add(t.v1.point);
			m.add(t.v2.point);
			m.add(t.v3.point);
		}
		ij3d.Image3DUniverse univ = new ij3d.Image3DUniverse();
		univ.show();
		ij3d.Content c = univ.addTriangleMesh(m, null, bla);
		c.setShaded(false);
		System.out.println(m);
	}
}

